// *****************************************************************************
// *
// *    Routines for communication with serial programmer
// *
// *****************************************************************************
//#define WIN32_LEAN_AND_MEAN
#include <conio.h>
#include <stdio.h>
#include <stdint.h>
#include <windows.h>
#include <string.h>
#include <time.h>
#include "Comport.h"
#include "target.h"
#include "SPRGcom.h"

extern int logflag;
extern FILE *logfile;
//int writereturn = 0;
int fullduplexflag = 0; 
int comchksum;
char txbuffer[TXBUFSIZE];
int tx_count = 0;

// *****************************************************************************
// write byte transfer into logfile
void logwrite(int data, int dir)
{
	static int current_dir = 0;
	static int count = 0;
		
	if (logflag == 0)
		return;				// logging is off
		
	if((dir != current_dir) || (++count > 16))	// data direction change or time for a newline?
	{
		current_dir = dir;
		count = 0;
		if(dir != 0)		// != 0 means "Tx"
			fprintf(logfile,"\nTx:");
		else
			fprintf(logfile,"\nRx:");		
	}
	fprintf(logfile," %02x", (unsigned char) data);
}

// *****************************************************************************
// transmit bytes from txbuffer, if tx buffer is not empty
void flushTxBuffer(void)
{
	int i;
	char inchar;
	
	if(tx_count)
	{
		ComWrite(txbuffer, tx_count);
		if(fullduplexflag == 0)			// SW-Half Duplex:
		{
			for (i=0; i<tx_count; i++)	// read back echo of transmitted data
			{
				inchar = rxWait(10);
				if(txbuffer[i]-inchar)
				{
					printf("\ncommunication error!\n");
					exit(EXIT_FAILURE);
				}
			}
		}
		tx_count = 0;
	}
}

// *****************************************************************************
// Write char into txbuffer, flush if tx buffer full
void WriteTxBuffer(char txdata)
{
	txbuffer[tx_count++] = txdata;
	if(tx_count >= TXBUFSIZE)
		flushTxBuffer();
}

// *****************************************************************************
// Receive byte from comport.
// Returns received data or -1 on timeout. 
int rxWait(int timeout)
{
	int d, timecnt = 0;
	
	while(ComGetReadCount() == 0)
//	while((d = ComRead()) < 0)
	{
		if(++timecnt > timeout)			// timeout?.
		{
			if(logflag)	fprintf(logfile," !!! Timeout !!!");
			return -1;
		}
		Sleep(5);
	}
	return ComRead();
}

// *****************************************************************************
// Receive byte from comport.
// Returns received data or -1 on timeout. 
int rxByteLog(void)
{
	flushTxBuffer();
	int d = rxWait(200);
	comchksum += d;
	if(d >= 0)
		logwrite(d,0);
	return d;
}

// *****************************************************************************
// Receive 2 bytes = 1 PIC data word (14 bits) from comport.
// Returns received data or -1 on timeout. 
int rxWord (void)
{
	int msb, lsb;
	
	lsb = rxByteLog();
	if(lsb < 0)
		return -1;
	msb = rxByteLog();
	if(msb < 0)
		return -1;
	return ((msb << 8) | lsb);	
}

int getchksum (void)
{
	if(rxByteLog() < 0)
		return -1;
	else
		return (comchksum & 0xFF);
}

// *****************************************************************************
// sends single byte command or cmd+data from buffer to programmer device.

void send2prg(char data)
{
	comchksum = 0;							// initialize checksum for replied data
	while(ComGetReadCount())
	{
		if(ComRead() == NAK)				// flush Rx-buffer
		{
			logwrite(NAK,0);
			exit(EXIT_FAILURE);
		}
	}
	WriteTxBuffer(data);
	WriteTxBuffer(0x100-data);
	logwrite(data,1);
}
  
void send2prg(char *buffer, int count)
{
	int i, j, retval, checksum = 0;
	unsigned char c;		

	comchksum = 0;							// initialize checksum for replied data

	while(ComGetReadCount())
	{
		if(ComRead() == NAK)				// flush Rx-buffer
		{
			logwrite(NAK,0);
			exit(EXIT_FAILURE);
		}
	}

	for(i=0;i<count;i++)
	{					
		WriteTxBuffer(buffer[i]);			// data -> txbuffer
		logwrite(buffer[i],1);
		checksum -= buffer[i];				// calculate checksum
	}
	WriteTxBuffer(checksum & 0xFF);
}


// *****************************************************************************
// connect with programmer device 

int connectPrg (void)
{
	int i, ch;
	int timestamp;
	
	// Test it communication line is single wire/half duplex or two wire/full duplex:
	while(ComGetReadCount())
		ComRead();				// flush Rx-buffer
		
	ComWrite(CMD_GETSTAT);		// send "read Status" command
	Sleep(20);
	if(ComRead() != CMD_GETSTAT)
		fullduplexflag = 1;		// if char was not echoed back: assume full duplex connection
	else
		fullduplexflag = 0;
	ComWrite(0x100-CMD_GETSTAT);	// (checksum, so communication is correct)
	Sleep(20);
	if(fullduplexflag == 0) 
		ComRead();				// read back echo (single wire, half duplex)
	return ComRead();			// return status byte from programmer
}




