//*****************************************************************************
//*
//*
//*		hexread.cpp
//*
//*
//*****************************************************************************
//
#include <stdio.h>
#include <stdint.h>
#include <ctype.h>
#include "hexread.h"
#include "target.h"
#include "memwrite.h"


FILE *source;
int checksum, minAddr, maxAddr;

//***************************************************************************** 
// converts two ASCII hex characters from file (source), 
// accumulates data in checksum
// returns data if byte could be read from file, -1 otherwise  

int ascii2byte(void)
{
	int result=0, ch, i;
	
	for(i=0; i<2; i++)
	{
		result <<= 4;
		ch = toupper(fgetc(source));
		if (isxdigit(ch))
		{
			if (isdigit(ch))
				result |= ch-'0';
			else
				result |= ch + 10 -'A';
		}
		else 
			return -1;
	}
	
	checksum = (checksum + result) & 0xFF;
	return result;
}
	
//***************************************************************************** 
// reads hex file and fills "picflash" with it's data
// returns -1 on error

int readHexFile(FILE *inputfile,FLASH *picflash, int flashsize)
{
	maxAddr = -1;
	minAddr = 0xFFFF;
	int i, length, rectype, state=0;
	char idata;
	int xaddr=0,addr,data8, wordaddr;
	
	source = inputfile;
	
	while(1)
	{
		while((idata = fgetc(source)) != ':')			// ignore all before ':'
			if(idata == EOF) return 1;					// end of file - unexpected!
		checksum = 0;
			
		if((length = ascii2byte()) <0)
			return -1; 
			
		if((data8 = ascii2byte()) <0)
			return -1;
		if((addr = ascii2byte()) <0)
			return -1;
		addr |= (xaddr << 16) | (data8 << 8);
			
		if((rectype = ascii2byte()) <0)
			return -1;
		
		if(picdata.type == T12F629)			// old midrange PICs: remap address ranges
		{
			if (addr >= 0x4200)				// > wordaddress 0x2100:
				addr += 0x1E000-0x4200;		// -> wordaddress 0xF000 (EEPROM area)
			else if (addr >= 0x4000)		// > wordaddress 0x2000:
				addr += 0x10000-0x4000;		// -> wordaddress 0x8000 (Config)
		}
			
		switch(rectype)
		{
		case 0:						// data record
			for(i=0;i<length;i++)
			{
				wordaddr = (addr+i)/2;		// hex file uses byte address, adress/2 = word address
				
				if(wordaddr >= MAXFLASHSIZE)
				{
					printf("\nHexfile contains invalid address!\n");
					return -1;
				}
				
				if((data8 = ascii2byte()) <0)
					return -1;
					
				if (maxAddr < wordaddr)				//  new maximum address:
					maxAddr = wordaddr;				// adjust max. used memory address
				if (minAddr > wordaddr)				//  new minimum address:
					minAddr = wordaddr;				// adjust min. used memory address
												
				if((wordaddr >= flashsize) && (wordaddr < 0x8000))
				{
					printf("\nData from hexfile exceed flashsize!\n");
					return -1;
				}
				if(wordaddr < flashsize)
					progflag |= PRG_PRG;
				if(wordaddr >= 0xF000)
					progflag |= PRG_DAT;
				if((wordaddr >= 0x8000) && (wordaddr <= 0x8003))
					progflag |= PRG_UID;
				if((wordaddr >= 0x8007) && (wordaddr <= 0x8008))
					progflag |= PRG_CFG;
					
				if ((addr+i) & 1)		// odd addresses are MSB
					picflash[wordaddr] = ((picflash[wordaddr] & 0xFF) | (data8 << 8)) & 0x3FFF;
				else
					picflash[wordaddr] = (picflash[wordaddr] & 0x3F00) | data8;
			}
			break;
		
		case 4:						// extended address
			if((data8 = ascii2byte()) <0)
				return -1;
			if((xaddr = ascii2byte()) <0)
				return -1;
			xaddr |= data8 << 8;
			break;
		
		case 1:						// end record
			break;
		}

		if(ascii2byte() <0)		// read checksum of record
			return -1;
		if(checksum)
			return -1;
			
		if(rectype == 1)		// if last record type == end
			break;				// exit from while-loop 
	}
	
	return 0;
}
