// *****************************************************************************
// *
// *     program target memory functions
// *
// *****************************************************************************



#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include "Options.h"
#include "SPRGcom.h"
#include "target.h"
#include "memwrite.h"

int progflag;
FLASH save_osccal, save_config;

// *****************************************************************************
// 	program all memory sections

void programming(void)
{	
	int i,j, temp;
	FLASH rwdata[MAXROWSIZE];

	
	if(progflag == 0)
		return;					//nothing to do
		
	if(picdata.flashsize & 0x8000)
	{
		printf("\nPIC not found in data base - Writing disabled!\n");
		exit(EXIT_FAILURE);
	}

	if(picdata.type == T12F629)
	{
		if(save_osccal > 0x3FFF)							// unless value was given by "fill" 
			save_osccal = readWord (picdata.osccal_loc);	// save osccal calibration data
		save_config = readWord (0x8007);					// save calibration bits from config reg.
		
		if ((progflag & (PRG_PRG|PRG_CFG|PRG_UID|PRG_ERA)) != 0)		// anything except data EEPROM to program?
		{
			progflag |= (PRG_PRG|PRG_CFG|PRG_UID);				// y: program all after erase
	
			if((progflag & PRG_ERA) == 0)				// skip if bulk erase is intended
			{	
				//------------- replace unchanged picflash[] by data from PIC's memory 
				if((save_config & 0x80) != 0)				// if code protection is off:
				{
					i=0;
					while(i < picdata.flashsize)
					{
						readWords (i, MAXROWSIZE, rwdata);
						for(j=0; j<MAXROWSIZE; j++)
						{
							if((picflash[i] > 0x3FFF)		// memory in it's init state
							&& (rwdata[j] < 0x3FFF))		// && flash not empty
								picflash[i] = rwdata[j];	// copy original data
							i++;
						}
					}
				}
				
				i=0x8000;
				readWords (i, 4, rwdata);			// user-ID
				for(j=0; j<4; j++)
					if(picflash[i+j] > 0x3FFF) picflash[i+j] = rwdata[j];
				if(picflash[0x8007] >= 0x3FFF)
					picflash[0x8007] = readWord (0x8007); 
			}
		}
	}
		 		
	if((picdata.osccal_loc != 0) 
	&& ((progflag & (PRG_PRG|PRG_CFG|PRG_ERA)) != 0))
	{
		picflash[picdata.osccal_loc] = save_osccal;
		progflag |= PRG_PRG;
	}
	
	if((picdata.save_configbits != 0) && (progflag & (PRG_CFG|PRG_ERA)))
	{
		save_config &= picdata.save_configbits;			// mask bits to save
		picflash[0x8007] &= 0x3FFF & ~picdata.save_configbits;	// clear masked bits here 
		picflash[0x8007] |= save_config;	// set bits to saved value
		progflag |= PRG_CFG;
	}
		
	if (((progflag & PRG_ERA) != 0) && (picdata.ee_size > 0))
	{
		send2prg(CMD_BLKERAEE);						// erase EEPROM
		sendDelayCmd (picdata.bulk_erasetime);
	}

	if(progflag & PRG_CFG)							// if config words will be written
	{
		progflag |= PRG_ERA;						// needs bulk erasing 
	}

	if ((picdata.type == T12F629)
		&& (progflag & (PRG_PRG|PRG_CFG|PRG_UID|PRG_ERA)))	// program, erase or config: 
		{					
			if ((save_osccal & 0x3C00) != 0x3400) 	// != retlw instruction
			{
				printf("\nimproper oscillator calibration instruction!\n");
				exit(EXIT_FAILURE);			
			}
			
			progflag |= (PRG_PRG|PRG_CFG|PRG_UID|PRG_ERA);
		}
	
	if((progflag & PRG_ERA) != 0)
	{
		printf("\nBulk Erasing");

//		int temp = readWord (0x8007);		
		setAddr (0x8000);					// erase all, including user-id
		send2prg(CMD_BLKERAPM);
		sendDelayCmd (picdata.bulk_erasetime);
		sendRdyAck();
	}

	if(progflag & PRG_PRG)
		writeProgramFlash();
	
	// ------------ write Data EEPROM:
	if(picdata.ee_size > 0)
	{
		temp = 0xFFFF;
		char command[3] = {CMD_LOADEEDATA,0,0};
	
		for(i=0; i<picdata.ee_size; i++)
			temp &= picflash[0xF000+i];
		if(temp <= 0x3FFF)							// any data there?
		{
			printf("\nProgramming Data EEPROM...");
//				send2prg(CMD_BLKERAEE);					// bulk erase data EEPROM
//				sendDelayCmd (picdata.bulk_erasetime);
			
			setAddr(0xF000);
			for(i=0; i<picdata.ee_size; i++)
			{
				if((i & 0x0F) == 0)					// avoid buffer overflow:
					sendRdyAck();					// pause every 32 loops
				if(picflash[0xF000+i] <= 0x3FFF)	// skip unwritten memory cells
				{
					command[1] = picflash[0xF000+i] & 0xFF;
					send2prg(command,3);
					send2prg(CMD_BEGINTPRG);
					sendDelayCmd (picdata.cfg_progtime);
				}	
				incAddr(1);
			}
			printf("\nVerifying Data EEPROM...");
			
			i=0xF000;
			do {
				readWords (0xF000+i, 32, rwdata);
				for(j=0; j<32; j++)
				{
					if((picflash[i+j] <= 0x3FFF) && ((picflash[i+j] ^ rwdata[j]) & 0xFF)) 
					{
						printf("\nVerify Error!\n");
						exit(EXIT_FAILURE);
					}
				}
				i += 32;
			} while(i< 0xF000 + picdata.ee_size);
		}
	}

	// ------------ write user ID:
	if(progflag & PRG_UID)							// any user-ID to program?
	{
		printf("\nProgramming User ID");

		setAddr (0x8000);
		if((progflag & PRG_ERA) == 0)				// if bulk erase was not performed
		{
			send2prg(CMD_ROWERA);					// (erases user-ID only)
			sendDelayCmd (picdata.row_erasetime);
		}
			
		for(j=0; j<4; j++)
		{
			sendLoadPMDCmd (picflash[0x8000+j]);
			send2prg(CMD_BEGINTPRG);					// single word programming
			sendDelayCmd (picdata.cfg_progtime);
			send2prg(CMD_READPMDATA);
			temp = rxWord();
			if((temp < 0) || ((picflash[0x8000+j] ^ temp) & 0x3FFF) != 0)
			{
				printf("\nVerify Error!\n");
				exit(EXIT_FAILURE);
			}
			incAddr(1);
		}
	}
	
	// ------------ write Config words:
	if((progflag & PRG_CFG) != 0)
	{
		printf("\nProgramming Config Words");
		setAddr (0x8007);
		
		if(picdata.type == T12F629)
		{
			int cfgwd = picflash[0x8007] & picdata.cfg_mask;
			sendLoadPMDCmd (cfgwd);
			send2prg(CMD_BEGINTPRG);					// single word programming
			sendDelayCmd (picdata.cfg_progtime);
			send2prg(CMD_READPMDATA);
			temp = rxWord();
			if((temp < 0) || ((cfgwd ^ temp) & 0x3FFF) != 0)
			{
				printf("\nError!\n");
				exit(EXIT_FAILURE);
			}
		}
		else
		{
			for(j=0; j<2; j++)
			{
				sendLoadPMDCmd (picflash[0x8007+j]);
				send2prg(CMD_BEGINTPRG);					// single word programming
				sendDelayCmd (picdata.cfg_progtime);
				send2prg(CMD_READPMDATA);
				temp = rxWord();
				if((temp < 0) || ((picflash[0x8007+j] ^ temp) & 0x3FFF) != 0)
				{
					printf("\nError!\n");
					exit(EXIT_FAILURE);
				}
				incAddr(1);
			}
		}
	}
	
	printf("\nOk!\n");
//		resetflag = 1;
}


// *****************************************************************************
// 	write program memory

void writeProgramFlash(void)
{
	int i, j, temp, rwsize;
	FLASH rwdata[MAXROWSIZE];
	
	rwsize = picdata.rowsize;					

	printf("\nProgramming program flash...");

	if(picdata.rowsize > 0)								// flash organized in erase blocks
	{
		for(i=0; i<picdata.flashsize; i+=rwsize)			// loop row by row...
		{
			temp = 0xFFFF;
			for(j=0; j<rwsize; j++)
				temp &= picflash[i+j];						// MSB = 0 if any word of row contains data
				
			if((temp & 0x8000) == 0)						// skip if whole row is empty
			{
				setAddr(i);
				if((progflag & PRG_ERA) == 0)				// erase row only if bulk erase was not performed
				{
					send2prg(CMD_ROWERA);
					sendDelayCmd (picdata.row_erasetime);
				}
				for(j=0; j<rwsize; j++)						// write latches (1 row)
				{
					sendLoadPMDCmd (picflash[i+j]);
					if((j+1) < rwsize) 
						incAddr(1);							// not yet end of row: addr++
				}
				send2prg(CMD_BEGINTPRG);					// program latches into flash
				sendDelayCmd (picdata.pm_progtime);			// (delay length of pp)
				sendRdyAck();								// wait while programmer busy
			}	
		}
	}
	else											// no block structure: single word programming
	{
		rwsize = MAXROWSIZE;								
		for(i=0; i < picdata.flashsize; i++)			
		{
			if(picflash[i] <= 0x3FFF)
			{
				setAddr(i);
				sendLoadPMDCmd (picflash[i]);	
				send2prg(CMD_BEGINTPRG);
				sendDelayCmd (picdata.pm_progtime);			// (delay length of pp)
			}
			if((i & 0x1F) == 0)						// every 32 loops:
				sendRdyAck();						// wait while programmer busy			
		}
	}
	
	// ------------ verify program memory:
	printf("\nVerifying program flash...");
	for(i=0; i<picdata.flashsize; i+=rwsize)			// loop row by row...
	{
		temp = 0xFFFF;
		for(j=0; j<rwsize; j++)
			temp &= picflash[i+j];						// MSB = 0 if any word of row contains data
			
		if((temp & 0x8000) == 0)						// skip if whole row is empty
		{
			setAddr(i);
			readWords (i, rwsize, rwdata);
			for(j=0; j<rwsize; j++)						// verify content
			{
				if(((picflash[i+j] ^ rwdata[j]) & 0x3FFF) != 0) 
				{
					printf("\nVerify Error!\n");
					exit(EXIT_FAILURE);
				}
			}
		}	
	}
}

//********************************************************************
// read memory row (size= flash erase block) program flash into picflash[address]
void readMemRow (int address)
{
	int i;
	if(address < picdata.flashsize)				// (program flash)
	{
		address &= ~(picdata.rowsize-1);		// = first address of row
		readWords (address, picdata.rowsize, &picflash[address]);
	}
}



// *****************************************************************************
// save contents of rows that partially extend into given range 
void saveCutRows(int from, int to)
{
	
	if((picdata.rowsize != 0) && (from <= picdata.flashsize))
	{
		if ((from/picdata.rowsize*picdata.rowsize) != from)		// row boundary?
			readMemRow(from);									// no, save contents row for partial erase
		if ((to | picdata.rowsize-1) != to)						// row boundary?
			readMemRow(to);										// no, save contents row for partial erase
	}
	if(from <= picdata.flashsize)
		progflag |= PRG_PRG;					// program flash section
	if(to >= 0xF000)
		progflag |= PRG_DAT;					// eeprom section
	if((from <= 0x8003) && (to >= 0x8000))	
		progflag |= PRG_UID;
	if((from <= 0x8008) && (to >= 0x8007))	
		progflag |= PRG_CFG;

}



