// *****************************************************************************
// *
// *     target access
// *
// *****************************************************************************

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include "SPRGcom.h"
#include "target.h"

FLASH picflash[MAXFLASHSIZE];
picdat_t picdata;
int addrcnt;
FILE *picdb;
char picname[20];  
extern char progname[];
extern char progpath[];

#define is_goto(opcode) ((opcode & 0x3800) == 0x2800)
#define is_bra(opcode) ((opcode & 0x3E00) == 0x3200)	
#define is_movlp(opcode) ((opcode  & 0x3F80) == 0x3180)
#define is_skipifnotwdto(opcode) (opcode == 0x1E03)


// *****************************************************************************
// try to set properties from PIC data file

void SetPICproperties(void)
{
	char string[256];
	char name[20];
	char c, op;
	int dev_id, mask, rowsize, flashsize;
	int parval;
	
	strcpy(string,progpath);
	strcat(string,".dat");

	// init to defaults...
	picdata.flashsize = 0x8000;					// max. size for unknown PIC	
	picdata.rowsize = 0x10;
	picdata.bulk_erasetime = 5000;
	picdata.cfg_progtime = 5000;
	picdata.ee_size = 0;
	picdata.pm_progtime = 2500;
	picdata.row_erasetime = 2500;
	picdata.vpp = 85;
	picdata.type = T12F1840;					// default: enhanced midrange (e.g. 12F1840)
	picdata.osccal_loc = 0;
	picdata.save_configbits = 0;
	picdata.cfg_mask = 0x3FFF;
	
	picdb = fopen(string, "r");
	if(picdb == NULL)
	{
		picdb = fopen(string, "w");
		if(picdb != NULL)
		{
			fprintf(picdb, "* PIC data base for %s.exe\n",progname);
			fprintf(picdb, "*\n* Line format:\n");
			fprintf(picdb, "* <device name> <hardware ID> <mask hardware ID> <erase blk size> <prog flash size> [optional parameters]\n");
			fprintf(picdb, "* (all above numbers: hexadecimal, size=words)\n");
			fprintf(picdb, "* optional parameters (default val.), numbers: decimal or hexadecimal\n");
			fprintf(picdb, "* B:<erasetime>        (5000) bulk erase time (microseconds)\n");
			fprintf(picdb, "* D:<prgtime>          (5000) programming pulse width for EEPROM/Config Mem. (microseconds)\n");
			fprintf(picdb, "* E:<EEPROM size>      (0) size of data EEPROM\n");
			fprintf(picdb, "* O:<address>          (0) OSCCAL memory address (e.g.for PIC12F629: 0x3FF)\n");
			fprintf(picdb, "* P:<prgtime>          (2500) programming pulse width for program flash (microseconds)\n");
			fprintf(picdb, "* R:<erasetime>        (2500) row erase time (microseconds)\n");
			fprintf(picdb, "* S:<bitmask>          (0) bits in Config word to save & restore (e.g. BG0,BG1)\n");		
			fprintf(picdb, "* T:<num>              (0) typecode for selecting special behaviour\n");
			fprintf(picdb, "* V:<volts>            (85) programming voltage (1/10 Volts)\n");
			fprintf(picdb, "* Z:<mask>             (0x3FFF) bit clr mask for hardware zeros (config-wd.)\n");
			
			fprintf(picdb, "PIC12F1840 0x1B80 0x3FE0 0x0020 0x1000 E:256\n");
			fprintf(picdb, "PIC12LF1840 0x1BC0 0x3FE0 0x0020 0x1000 E:256\n");
			fprintf(picdb, "PIC16F1503 0x2CE0 0x3FE0 0x0010 0x0800\n");
			fprintf(picdb, "PIC16LF1503 0x2DA0 0x3FE0 0x0010 0x0800\n");
			fprintf(picdb, "PIC16F1705 0x3055 0x3FFF 0x0020 0x2000\n");
			fprintf(picdb, "PIC16LF1705 0x3057 0x3FFF 0x0020 0x2000\n");
			fprintf(picdb, "PIC16F1825 0x2760 0x3FE0 0x0020 0x2000 E:256\n");
			fprintf(picdb, "PIC16LF1825 0x2860 0x3FE0 0x0020 0x2000 E:256\n");
			fprintf(picdb, "PIC12F629 0x0F80 0x3FE0 0x0000 0x400 E:128  T:1 V:95 D:6000 B:8000 O:0x3FF S:0x3000 Z:0x31FF\n");
			fprintf(picdb, "PIC12F675 0x0FC0 0x3FE0 0x0000 0x400 E:128  T:1 V:95 D:6000 B:8000 O:0x3FF S:0x3000 Z:0x31FF\n");
			fprintf(picdb, "PIC16F630 0x10C0 0x3FE0 0x0000 0x400 E:128  T:1 V:95 D:6000 B:8000 O:0x3FF S:0x3000 Z:0x31FF\n");
			fprintf(picdb, "PIC16F676 0x10D0 0x3FE0 0x0000 0x400 E:128  T:1 V:95 D:6000 B:8000 O:0x3FF S:0x3000 Z:0x31FF\n");
			fclose(picdb);
			printf("\nNew file (%s.dat) written\n",progname);
			picdb = fopen(string, "r");
		}
	}
	
	if (picdb != NULL)	
	{
		while ((c = fgetc(picdb)) != EOF)
		{
			while(isspace(c))
				c = fgetc(picdb);			// ignore spaces at start of line
			if(c == 'P')					// ignore line if it does not begin with 'P'
			{
				ungetc('P',picdb);
				if(((fscanf(picdb,"%20s %x %x %x %x", name, &dev_id, &mask, &rowsize, &flashsize)) == 5)
				&& (((picdata.hardware_id ^ dev_id) & mask) == 0))
				{
					strcpy(picdata.picname,name);
					picdata.rowsize = rowsize;	
					picdata.flashsize = flashsize;	
					
					while((c = fgetc(picdb)) != '\n')
					{
						if(c == EOF)
							break;
						if(c == ':')
						{
							if(fscanf(picdb,"%s", string) != 1)
								break;	
							parval = dechextoi(string);
							switch (op)
							{
							case 'B':
								picdata.bulk_erasetime = parval;
								break;	
							case 'D':
								picdata.cfg_progtime = parval;
								break;	
							case 'E':
								if(parval != 0)
								{
									picdata.ee_size = parval;
									picdata.eeaddr = 0xF000;
								}
								break;	
							case 'O':
								picdata.osccal_loc = parval;
								break;	
							case 'P':
								picdata.pm_progtime = parval;
								break;	
							case 'R':
								picdata.row_erasetime = parval;
								break;	
							case 'S': 
								picdata.save_configbits = parval;
								break;
							case 'T':
								picdata.type = parval;
								break;	
							case 'V':
								picdata.vpp = parval;
								break;	
							case 'Z':
								picdata.cfg_mask = parval;
								break;	
							}
						}
						else
							op = c;
					}
					fclose(picdb);
					return;
				}
			}
			
			while((c = fgetc(picdb)) != '\n')	// ignore rest of line
				if(c == EOF) break;

		}
		fclose(picdb);
	}
}


// *****************************************************************************
// connect with programmer and get PIC properties
// returns -1 in case of no connection
// else returns status byte of programmer device

int GetTargetInfo (void)
{
	int pstat, i, addr=0;
	int opcode;

	if ((pstat=connectPrg()) < 0)		// set connection type (full|half duplex)
	{
		printf("\nno Programmer connected!\n");
		exit(EXIT_FAILURE);
	}
	if((pstat & PSTAT_CONERR) != 0)
	{
		printf("\nICSP connection failure!\n");
		exit(EXIT_FAILURE);
	}
	sendSetVpp (85);				// set VPP = default value (8.5V)
	send2prg(CMD_PRGON);
	flushTxBuffer();
	if((i= rxWait(200)) != ACK)
	{
		printf("\nVPP activation failed!\n");
		exit(EXIT_FAILURE);
	}
	addrcnt = 0;
	readWords (0x8000, 4, picdata.user_idmem);
	picdata.user_id = 0;
	for(i=0; i<4; i++)
	{
		picdata.user_id <<= 4;
		picdata.user_id |= picdata.user_idmem[i] & 0x0F;
	}	
	picdata.hardware_id = readWord (0x8006);
	
	if((picdata.hardware_id == 0) || picdata.hardware_id >= 0x3FFF)
	{
		printf("\nno target detected!\n");
		exit(EXIT_FAILURE);
	}
	SetPICproperties();
	sendSetVpp (picdata.vpp);				// set VPP = value from PIC data

	if(picdata.rowsize > MAXROWSIZE)
	{
		printf("\nunsupported rowsize!\n");
		exit(EXIT_FAILURE);
	}

	return pstat;
}

//********************************************************************
// check for valid memory range
int validMem (int address)
{
	int rw;
	if((picdata.ee_size != 0)
	&& (address >= picdata.eeaddr)
	&& (address < picdata.eeaddr+picdata.ee_size))	// within eeprom range
		return 1;	// memory contains a data byte

	if((address < 0x8000) && (address >= picdata.flashsize))		// above program flash, but not config
		return 0;	// no memory there
	
	if((rw = picdata.rowsize) == 0)			// e.g. P12F629
		rw = 8;
		
	if(address >= (0x8000+rw))	// above config
		return 0;
	return 2;		// memory contains a data word
}

//********************************************************************
// read single word from PIC into memory 
FLASH readWord (int address)
{
	FLASH retv;
	readWords (address, 1, &retv);
	return retv;
}

//********************************************************************
// read multiple words from PIC into memory 
void readWords (int address, uint8_t count, FLASH *destination)
{
	setAddr (address);
	addrcnt += count-1;					// repeated cmd will leave internal address = last read
	sendRptCmd (count);
	if(address >= 0xF000)				// depending on address range:
		send2prg(CMD_READEEDATA);		// read data EEPROM
	else
		send2prg(CMD_READPMDATA);		// or read program|config memory
		
	while(count>0)
	{
		*destination++ =  rxWord ();
		count--;
	}
	if(getchksum() != 0)
	{
		printf("\ncommunication error!\n");
		exit(EXIT_FAILURE);
	}
}

// *****************************************************************************
// send CMD_VPPVAL
void sendSetVpp (int vppval)
{
	char command[2];
	
	command[0] = CMD_VPPVAL;
	command[1] = vppval;
	send2prg(command,2);
}


// *****************************************************************************
// send repeat command to programmer
void sendRptCmd (int rptcnt)
{
	char command[2];
	
	command[0] = CMD_REPEAT;
	command[1] = rptcnt;
	if(rptcnt > 1)
		send2prg(command,2);
}

// *****************************************************************************
// send CMD_RDYSNS, wait for ACK
void sendRdyAck (void)
{
	send2prg(CMD_RDYSNS);
	if(rxByteLog() != ACK)
	{
		printf("\ncommunication error!\n");
		exit(EXIT_FAILURE);
	}
}

// *****************************************************************************
// send wait command to programmer
void sendDelayCmd (int delay)
{
	char command[3];
	
	command[0] = CMD_DELAY;
	command[1] = delay & 0xFF;
	command[2] = delay >> 8;
	send2prg(command,3);
}

// *****************************************************************************
// send Load Data for Program Memory
void sendLoadPMDCmd (int data)
{
	char command[3];
	
	data &= 0x3FFF;
	command[0] = CMD_LOADPMDATA;
	command[1] = data & 0xFF;
	command[2] = data >> 8;
	send2prg(command,3);
}

// *****************************************************************************
// increment address counter cnt times
void incAddr (int cnt) 
{
	int sensflag = 0;
	int step;
	addrcnt += cnt;
	
	while (cnt > 0)
	{
		step = cnt;
		if (step > 255)		//  > max.repeat number?
			step = 255;		// y: limit to max. value
		
		if(step > 1)	
			sendRptCmd (step);
		send2prg(CMD_INCADDR);
		cnt -= step;
		if(++sensflag > 10)
		{
			sensflag=0;
			sendRdyAck();
		}
	}
}

// *****************************************************************************
// Perform a reset of PICs address counter and clear addrcnt

void sendResetAddr (void)
{
	addrcnt = 0;
	if(picdata.type == T12F629)		
	{	
		send2prg(CMD_PRGOFF);		// CMD_RESADDR command doe not exist!
		send2prg(CMD_PRGON);		// reset address by exit/enter prog.mode
		flushTxBuffer();
		if(rxWait(200) != ACK)
		{
			printf("\nVPP activation failed!\n");
			exit(EXIT_FAILURE);
		}			
	}
	else
	{
		send2prg(CMD_RESADDR);
	}
}

// *****************************************************************************
// set address counter to desired address
// 
void setAddr (int newaddr)
{
	char cmd[3] = {CMD_LOADCONFIG,0,0};	// command sets address to config mem (0x8000)

	if(newaddr == 0)
	{
		sendResetAddr();
	}

	if(newaddr < 0x8000)				// new address in program memory
	{
		if(newaddr < addrcnt)
		{
			sendResetAddr();
		}
		incAddr (newaddr-addrcnt); 
	}
	else if(newaddr >= 0xF000)				// new address in EEPROM memory
	{
		if((newaddr < addrcnt) || (addrcnt < 0xF000))
		{
			sendResetAddr();
			addrcnt = 0xF000;
		}
		incAddr (newaddr-addrcnt); 
	}
	else									// new address is configuration
	{
		if((newaddr < addrcnt) || (addrcnt >= 0xF000) || (addrcnt < 0x8000))	
		{
//			sendResetAddr();
			addrcnt = 0x8000;
			send2prg(cmd,3);				// load configuration
		}
		incAddr (newaddr-addrcnt); 
	}
	
}


//***************************************************************************** 
// converts decimal or hexadecimal number from ASCII string to integer

int dechextoi(char *str)
{
	int retval = 0, i=0;
	while(str[i]=='0')				// skip leading zeros
		i++;
	if(isdigit(str[i]))
		return atoi(&str[i]);		// it's a decimal number!
	if(toupper(str[i]) == 'X')		// 'X' denotes hexadecimal radix
		i++;
	while(isxdigit(str[i]))
	{
		retval <<= 4;
		if (isdigit(str[i]))
			retval |= str[i]-'0';
		else
			retval |= toupper(str[i]) + 10 -'A';
		i++;
	}
	return retval;
}

