//*****************************************************************************
//*
//*		
//*		target.h
//*
//*
//*****************************************************************************
#ifndef 	__TARGET_H__
#define 	__TARGET_H__

#define MAXFLASHSIZE 0x10000
#define MAXROWSIZE 32			// max. flash erase block size (words)

// Definitions of PIC commands (MSBs used by programmer for easy command classification):
#define CMD_LOADCONFIG 0x80
#define CMD_LOADPMDATA 0x82
#define CMD_LOADEEDATA 0x83
#define CMD_READPMDATA 0x44
#define CMD_READEEDATA 0x45
#define CMD_INCADDR	0x06
#define CMD_RESADDR 0x16
#define CMD_BEGINTPRG 0x08
#define CMD_BEGEXTPRG 0x18
#define CMD_ENDEXTPRG 0x0A
#define CMD_BLKERAPM 0x09
#define CMD_BLKERAEE 0x0B
#define CMD_ROWERA 0x11

// Special programmer commands:
#define CMD_RDYSNS 0xC0
#define CMD_REPEAT 0xD5
#define CMD_VPPVAL 0xD4
#define CMD_PRGON 0xC3
#define CMD_PRGOFF 0xC2
#define CMD_GETSTAT 0xC1
#define CMD_DELAY 0xE6
#define CMD_RUN 0xC7
#define CMD_BOOTLDR 0xC8

// programmer status values:
#define PSTAT_CONERR 0x80			// ICSP cable connection error
#define PSTAT_TRGVCC 0x10			// VCC supplied by target

// Algorithm defined by example device name (picdata.type):
#define T12F1840 0
#define T12F629 1

typedef unsigned short int FLASH;
typedef unsigned char SFR;
 
typedef struct
{
	FLASH hardware_id;			// PIC's hardware ID word (type and revision)
	int user_id;				// user-ID, packed into 16 bit
	FLASH user_idmem[4];		// 4 words user ID memory
	char picname[20];			// pic device name
	int flashsize;				// flash memory size (words) of PIC
	int rowsize;				// PIC's erase block/row size (words)
	int eeaddr;
	int vpp;					// programming voltage
	int ee_size;				// size of EEPROM
	int pm_progtime;			// program memory prog.pulse width
	int cfg_progtime;			// config/data memory prog.pulse width
	int row_erasetime;			//
	int bulk_erasetime;			//
	int save_configbits;		// e.g. mask for BG0/BG1 calibration bits (12F675)
	int osccal_loc;				// memory location to store osccal-calibration value
	int debugflag;				// (for testing)
	int type;					// type specific algorithm select
	int cfg_mask;				// 0 bits denote unused bits in config reg.
} picdat_t;

int GetTargetInfo (void);
int CheckCompatibility(void);
void sendSetVpp (int vppval);
void setAddr (int newaddr);
void sendDelayCmd (int delay);
void incAddr (int cnt); 
void sendRptCmd (int rptcnt);
void sendLoadPMDCmd (int data);
void sendRdyAck (void);
FLASH readWord (int address);
int validMem (int address);
void readWords (int address, uint8_t count, FLASH *destination);
int dechextoi(char *str);

extern FLASH picflash[MAXFLASHSIZE];
extern FLASH resetvect[2];
extern picdat_t picdata;
extern int addrcnt;

#endif 	


